# Copyright (C) 2023 Wibo Kuipers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional

from .globals import CONFIG_TABLE_NAME
from .types import ValidationIssueLevel


class NoConnection(Exception):
    def __init__(self):
        super().__init__("No database connection active.")


class DBConfigIncomplete(Exception):
    def __init__(self, missing: str):
        super().__init__(f'There seems to be an issue with the database config. It is initialized, but does not return a valid {missing}. Please manually check the "{CONFIG_TABLE_NAME}" table in your database.')


class NoDatabaseEngineSpecified(ValueError):
    def __init__(self):
        super().__init__("No database type was specified.")


class DBNotInitialized(Exception):
    def __init__(self):
        super().__init__("The database has not yet been initialized.")


class NoNetworkConnection(Exception):
    def __init__(self):
        super().__init__("No active internet connection was found.")


class MissingVersionData(ValueError):
    def __init__(self) -> None:
        super().__init__("A version source must be supplied for the update to run")


class IncompleteVersionData(ValueError):
    def __init__(self) -> None:
        super().__init__(
            'Version information data not complete. Must contain "latest", "version" and "name". Latest is the latest version number, version is a JSON object containing the database structure and name is the database template name.'
        )


class NeedsConfirmation(ValueError):
    def __init__(self) -> None:
        super().__init__("Did you forget to set confirm to True? This is a safety feature!")


class IncompleteVersionObject(ValueError):
    def __init__(self, key=None, object=None) -> None:
        if key == None or object == None:
            message = "Database version data is incomplete or broken."
        else:
            message = f'Database version data is incomplete or broken. "{object}" is missing key "{key}".'
        super().__init__(message)


class DBTemplateNoMatch(ValueError):
    def __init__(self) -> None:
        super().__init__("This database uses a different database version source. The template name does not match the one previously used to update this database.")


class IncompatibleColumnAttributes(ValueError):
    def __init__(self, *args, version: Optional[str] = None) -> None:
        attr_list = []
        for attr in args:
            attr_list.append(f'"{attr}"')

        message = ""

        if not version == None:
            message += f"{version}: "

        message += f'Column attributes {", ".join(attr_list)} are not compatible.'

        super().__init__(message)


class VersionSourceValueError(ValueError):
    def __init__(self, message: str):
        super().__init__(message)



# Copyright (C) 2023 Wibo Kuipers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from alphadb.utils.common import convert_version_number
from alphadb.utils.concatenate.column import concatenate_column, get_column_renames
from alphadb.utils.concatenate.primary_key import get_primary_key
from alphadb.utils.query.column.addcolumn import addcolumn
from alphadb.utils.query.column.modifycolumn import modifycolumn


def altertable(version_source: dict, table_name: str, version: str):
    query = f" ALTER TABLE {table_name}"

    #### Get the data for the current table
    table_data = next(v["altertable"][table_name] for v in version_source["version"] if v["_id"] == version)

    #### If primary key is changed, AUTO_INCREMENT must be removed from the column
    if "primary_key" in table_data:
        #### Creating the query for the primary key is done after all column modification
        #### Here we create a modifycolumn to remove the AUTO_INCREMENT from the old primary key column
        old_primary_key = get_primary_key(version_source["version"], table_name=table_name, before_version=version)

        if not old_primary_key == None:
            column_renames = get_column_renames(version_source["version"], column_name=old_primary_key, table_name=table_name, order="ASC")

            #### If the column is renamed, get historycal column name for version
            version_column_name = old_primary_key
            for rename in reversed(column_renames):
                if convert_version_number(version) >= rename["rename_version"]:
                    version_column_name = rename["new_name"]
                    break  ## If the name has been found, break out of the loop
                else:
                    version_column_name = old_primary_key

            if "modifycolumn" in table_data:
                if version_column_name in table_data["modifycolumn"]:
                    table_data["modifycolumn"][version_column_name]["a_i"] = False
                else:
                    table_data["modifycolumn"][version_column_name] = {"recreate": False, "a_i": False}
            else:
                table_data["modifycolumn"] = {version_column_name: {"recreate": False, "a_i": False}}

    if "dropcolumn" in table_data:
        for column in table_data["dropcolumn"]:
            query += f" DROP COLUMN {column}"
            query += ","

    if "addcolumn" in table_data:
        for column in table_data["addcolumn"]:
            partial = addcolumn(table_data, table_name=table_name, column_name=column, version=version)
            if partial == None:
                continue
            query += partial
            query += ","

    if "modifycolumn" in table_data:
        for column in table_data["modifycolumn"]:
            if "recreate" in table_data["modifycolumn"][column] and table_data["modifycolumn"][column]["recreate"] == False:
                table_data["modifycolumn"][column] = concatenate_column(version_source["version"], table_name=table_name, column_name=column)

            partial = modifycolumn(table_data, table_name=table_name, column_name=column, version=version)

            #### If column data is None, its some attribute that should be handled later (foreign_key, primary_key, etc...)
            if partial == None:
                continue
            query += partial
            query += ","

    if "renamecolumn" in table_data:
        for column in table_data["renamecolumn"]:
            query += f" RENAME COLUMN {column} TO {table_data['renamecolumn'][column]}"
            query += ","

    if "primary_key" in table_data:
        if table_data["primary_key"] == None:
            query += " DROP PRIMARY KEY"  ## Fails when AI, AI column must be key
            query += ","

        #### Change the primary key
        if table_data["primary_key"] in table_data:
            ...

    #### Remove trailing comma
    query = query[:-1]

    query += ";"

    return query



# Copyright (C) 2023 Wibo Kuipers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sys
from typing import TypedDict

import inquirer

from cli.utils.common import console


class MySQLCredentials(TypedDict):
    host: str
    user: str
    password: str
    database: str
    port: int


def check_port(val):
    "Check if port is a number"
    try:
        int(val if val else 3306)
    except ValueError:
        return False
    return True


def get_mysql_creds() -> MySQLCredentials:
    "Prompt user to input credentials"
    console.print(
        "[cyan]Connecting to a MySQL database requires you to provide log-in credentials[/cyan]\n"
    )
    console.print("[cyan]Host:[/cyan]", "URL/IP (Default localhost)")
    console.print("[cyan]User:[/cyan]", "User with permissions to alter the database")
    console.print("[cyan]Password:[/cyan]", "The users password")
    console.print("[cyan]Database:[/cyan]", "The name of the database to connect to")
    console.print("[cyan]Port:[/cyan]", "(Default 3306)\n\n")

    questions = [
        inquirer.Text("host", message="Host (localhost)"),
        inquirer.Text("user", message="User"),
        inquirer.Password("password", message="Password"),
        inquirer.Text("database", message="Database"),
        inquirer.Text(
            "port", message="Port (3306)", validate=lambda _, x: check_port(x)
        ),
    ]

    answers = inquirer.prompt(questions)

    #### If answers is None, user probably aborted
    if answers == None:
        sys.exit()  ## Exit because returning raises error as there is no credentials returned

    return {
        "host": answers["host"] if not answers["host"] == "" else "localhost",
        "user": answers["user"],
        "password": answers["password"],
        "database": answers["database"],
        "port": int(answers["port"]) if not answers["port"] == "" else 3306,
    }



# Copyright (C) 2023 Wibo Kuipers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Literal, Optional, Union

from cryptography.fernet import Fernet
from inquirer import Confirm, prompt
from mysql.connector import DatabaseError, InterfaceError

from alphadb import AlphaDB, VersionSourceVerification
from alphadb.utils.exceptions import DBConfigIncomplete, DBTemplateNoMatch, IncompleteVersionData, MissingVersionData
from alphadb.utils.types import ValidationIssueLevel
from cli.utils import globals
from cli.utils.common import console, print_title
from cli.utils.config import config_get, config_write
from cli.utils.connect import get_mysql_creds
from cli.utils.decorators import connection_check
from cli.utils.version_source import select_version_source, vs_path_to_json


def connect():
    "Connect to a database"

    print_title("connect")

    globals.db = AlphaDB()
    creds = get_mysql_creds()

    try:
        globals.db.connect(
            host=creds["host"],
            user=creds["user"],
            password=creds["password"],
            database=creds["database"],
            port=creds["port"],
        )

    except (DatabaseError, InterfaceError) as e:
        if hasattr(e, "msg"):
            console.print(f"\n[red]{e.msg}[/red]\n")
        else:
            console.print(f"\n[red]{e}[/red]\n")
        return

    pass_bytes = creds["password"].encode()
    f = Fernet(config_get("CONFIG", "secret"))
    pass_encrypted = f.encrypt(pass_bytes)

    config_write(
        {
            "DB_SESSION": {
                "host": creds["host"],
                "user": creds["user"],
                "password": pass_encrypted,
                "database": creds["database"],
                "port": creds["port"],
            }
        }
    )

    console.print(f'\n[green]Successfully connected to database:[/green] [cyan]"{creds["database"]}"[/cyan]\n')

    return


@connection_check()
def init():
    "Initialize database"

    #### Initialize loader
    with console.status("[cyan]Getting the database ready[/cyan]", spinner="bouncingBall") as _:
        init = globals.db.init()

    if init == "already-initialized":
        console.print("[yellow]The database is already initialized[/yellow]\n")
        return

    if init == True:
        console.print("[green]Database successfully initialized[/green]\n")
        return


@connection_check()
def status():
    "Get database status"
    print_title("database status")

    check = globals.db.status()

    print(f'Database: {check["name"]}')
    print(f'Template: {check["template"]}')
    if check["init"] == True:
        console.print("Status: ", "[cyan]Initialized[/cyan]")
    else:
        console.print("Status: ", "[yellow]Uninitialized[/yellow]")
    print(f'Version: {check["version"]}\n')

    return


@connection_check()
def update(nodata=False, verify: Optional[bool] = True, allowed_error_priority: Union[ValidationIssueLevel, Literal["ALL"]] = "LOW"):
    "Update database"

    print_title("update")

    #### Initialize loader
    with console.status("[cyan]Checking database[/cyan]", spinner="bouncingBall") as loader:
        #### Check database status
        status = globals.db.status()
        if status["init"] == False:
            console.print(f"[yellow]Database [cyan]{status['name']}[/cyan] has not yet been initialized[/yellow]\n")
            return

    try:
        version_source_path = select_version_source()

        #### If source path is None, user probably aborted
        if version_source_path == None:
            return

        with console.status("[cyan]Reading version source[/cyan]", spinner="bouncingBall") as loader:
            version_information = vs_path_to_json(version_source_path)

            #### If version information is None, an error has likely occured and we should not proceed
            if version_information == None:
                return

            loader.update("[cyan]Running updates on the database[/cyan]")

            update = globals.db.update(version_source=version_information, no_data=nodata, verify=verify, allowed_error_priority=allowed_error_priority)

            if update:
                if not type(update) == bool:
                    if update == "up-to-date":
                        console.print(f"[blue]Database is already the latest version [cyan]({status['version']})[/cyan][/blue]\n")
                        return

                    if update[:10] == "The databa":
                        console.print(f"[yellow]{update}[/yellow]\n")
                        return

                console.print("[green]Database successfully updated to the latest version[/green]\n")
            else:
                console.print("[red]An error occured[/red]\n")

    except (DBTemplateNoMatch, IncompleteVersionData, MissingVersionData, DBConfigIncomplete) as e:
        console.print(f"[red]{e}[/red]\n")


@connection_check()
def vacate(confirm=False):
    "Empty database"
    print_title("vacate")

    #### Only works when confirm==True is specified
    if not confirm == True:
        console.print(f"[yellow]The vacate function requires the[/yellow] [red]--confirm[/red] [yellow]option.[/yellow]")
        print("This is a safety feature that hopefully prevents unintended data loss.\n\nDon't worry, you'll still be prompted for confirmation!\n")
        return

    console.print(f"[yellow]The vacate function[/yellow] [red]deletes all data in the database[/red].")
    console.print("[yellow]This action can [red]NOT[/red] be undone.[/yellow]\n")

    #### Prompt user for confirmation
    questions = [Confirm("confirm", message="Are you absolutely sure you want to completely delete all data?")]
    answers = prompt(questions)

    #### If answers is None, user likely aborted
    if answers == None:
        return

    if answers["confirm"]:
        with console.status("[cyan]Removing all data from database[/cyan]", spinner="bouncingBall") as _:
            #### Empty db
            if globals.db.vacate(confirm=confirm):
                console.print("[green]The database had successfully been emptied.[/green]\n")
    else:
        console.print("[cyan]Not empying[/cyan]\n")
    return


def verify_version_source():
    print_title("verify version source")

    version_source_path = select_version_source()

    #### If version source path is None, user likely aborted
    if version_source_path == None:
        return

    version_source = vs_path_to_json(version_source_path)

    #### If version source is None, an error likely occured and we should not proceed
    if version_source == None:
        return

    verification = VersionSourceVerification(version_source)

    output = verification.verify()

    if output == True:
        console.print(f"[green]Version source at [blue]{version_source_path}[/blue] verified without errors[/green]\n")
        return

    console.print(f"Version source at [blue]{version_source_path}[/blue] has [red]{len(output)} errors[/red]\n\n")

    for issue in sorted(output):
        if ":" in issue[1]:
            issue_path, issue_text = issue[1].rsplit(": ", 1)
        else:
            issue_path = ""
            issue_text = issue[1]

        if issue[0] == "LOW":
            console.print("[black on white]LOW VULNERABILITY: [/black on white]", f"[cyan]{issue_path}[/cyan]", issue_text)

        if issue[0] == "HIGH":
            console.print("[white on yellow]HIGH VULNERABILITY: [/white on yellow]", f"[cyan]{issue_path}[/cyan]", f"[yellow]{issue_text}[/yellow]")

        if issue[0] == "CRITICAL":
            console.print("[white on red]CRITICAL: [/white on red]", f"[cyan]{issue_path}[/cyan]", f"[red]{issue_text}[/red]")

    console.line()



# Copyright (C) 2023 Wibo Kuipers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
from cli.utils.config import config_get
from rich.console import Console
console = Console()

"Prints out the programs title"
def print_title(title:str):
    console.print(f"[green]----- [white]{title.upper()}[/white] -----[/green]\n")

"Clear the terminal"
def clear():

    #### Clear terminal
    if os.name == "posix":
        os.system("clear")
    else:
        os.system("cls")

    db_session_data = config_get("DB_SESSION", fallback=True)
    if not db_session_data == None:
        console.print(f'[cyan]Connected to database[/cyan] {db_session_data["database"]} [cyan]on[/cyan] {db_session_data["host"]}\n')



import pytest

from src.alphadb.utils.common import convert_version_number


def test_version_number_converts_to_int():
    assert convert_version_number("1.0.201") == 10201
    assert convert_version_number("0.5.0") == 50

    with pytest.raises(ValueError):
        convert_version_number("11a")



# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Literal, Optional, Union, get_args

import typer

from alphadb.utils.types import ValidationIssueLevel
from cli import __app_name__, __version__, commands
from cli.utils.common import console

app = typer.Typer()


@app.command(help="Initialize the currently active database")
def init() -> None:
    commands.init()


@app.command(help="Show to status of the currently active database")
def status() -> None:
    commands.status()


@app.command(help="Update the database (requires a version source)")
def update(
    nodata: Optional[bool] = typer.Option(
        False,
        "--no-data",
        help="Update only to the database structure, but do not include default data",
    ),
    verify: Optional[bool] = typer.Option(
        True,
        "--verify",
        help="Wether the version source should be verified before updating",
    ),
    allowed_error_priority: Optional[str] = typer.Option("LOW", "--allowed-error-priority", help="Set the priority level for the verification to cancel the update process"),
) -> None:
    prio: Union[ValidationIssueLevel, Literal["ALL"]] = "LOW"
    if not allowed_error_priority == None:
        levels = [get_args(x)[0] if not x == "ALL" else "ALL" for x in ["ALL", *get_args(ValidationIssueLevel)]]
        prio = allowed_error_priority.upper()  ## LSP will error, but this is checked below

        if not prio in levels:
            console.print(f"[red]{prio} is not a valid priority for --allowed-error-priority[/red]\n")
            console.print("Supported priorities are:")
            for l in levels:
                console.print(f" - {l}")

            print("  ")
            return
    commands.update(nodata=nodata if not nodata == None else False, verify=verify, allowed_error_priority=prio)


@app.command(help="Irriversibally deletes ALL data in the database")
def vacate(
    confirm: Optional[bool] = typer.Option(
        None,
        "--confirm",
        help="Needs to be specified. This is a safety feature. Only with this option specified the vacate function will be called",
    )
) -> None:
    commands.vacate(confirm=confirm if not confirm == None else False)


@app.command(help="Connect to a new database")
def connect() -> None:
    commands.connect()


@app.command("verify", help="Verify version source")
def verify_version_source() -> None:
    commands.verify_version_source()


def version_callback(value: bool) -> None:
    if value:
        typer.echo(f"{__app_name__} {__version__}")
        raise typer.Exit()
    return


@app.callback()
def main(
    version: Optional[bool] = typer.Option(
        None,
        "--version",
        "-V",
        help="Show the application's version and exit.",
        callback=version_callback,
        is_eager=True,
    )
) -> None:
    return



# Copyright (C) 2023 Wibo Kuipers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from alphadb.utils.query.column.definecolumn import definecolumn, prepare_definecolumn_data


def addcolumn(table_data, table_name: str, column_name: str, version: str):
    column_data = prepare_definecolumn_data(table_name=table_name, column=column_name, table_data=table_data["addcolumn"], version=version)

    #### If column data is None, its some attribute that should be handled later (foreign_key, primary_key, etc...)
    if column_data == None:
        return None

    return " ADD" + definecolumn(
        column_name=column_name,
        column_type=table_data["addcolumn"][column_name]["type"],
        length=column_data["length"],
        null=column_data["null"],
        unique=column_data["unique"],
        default=column_data["default"],
        auto_increment=column_data["auto_increment"],
    )



# Copyright (C) 2023 Wibo Kuipers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Literal, Optional

from alphadb.utils.common import convert_version_number


def concatenate_column(version_list: list, table_name: str, column_name: str):
    """
    Concatenate al changes of a column into a new column definition
    """

    column = {}

    #### Recursively check for column renames
    rename_data = get_column_renames(version_list=version_list, column_name=column_name, table_name=table_name)

    #### Create new variable for column name to assign old names to if the column has been renamed
    version_column_name = column_name

    for version in version_list:
        v = convert_version_number(version["_id"])

        #### If the column is renamed, get historical column name for version
        for rename in reversed(rename_data):
            if v <= rename["rename_version"]:
                version_column_name = rename["old_name"]
                break  ## If the name has been found, break out of the loop
            else:
                version_column_name = column_name

        #### Create table
        if "createtable" in version:
            if table_name in version["createtable"]:
                if version_column_name in version["createtable"][table_name]:
                    for attr in version["createtable"][table_name][version_column_name]:
                        column[attr] = version["createtable"][table_name][version_column_name][attr]

        #### Alter table
        if "altertable" in version:
            if table_name in version["altertable"]:
                #### Modify column
                if "modifycolumn" in version["altertable"][table_name]:
                    if version_column_name in version["altertable"][table_name]["modifycolumn"]:
                        this_mod = version["altertable"][table_name]["modifycolumn"][version_column_name]

                        recreate = True if not "recreate" in this_mod or this_mod["recreate"] == True else False
                        if recreate:
                            column = {}
                        for attr in version["altertable"][table_name]["modifycolumn"][version_column_name]:
                            if attr == "recreate":
                                continue  ## Recreate is not an attribute but an instruction for the updater
                            column[attr] = version["altertable"][table_name]["modifycolumn"][version_column_name][attr]

                #### Drop column
                if "dropcolumn" in version["altertable"][table_name]:
                    if version_column_name in version["altertable"][table_name]["dropcolumn"]:
                        column = {}

                #### Add column
                if "addcolumn" in version["altertable"][table_name]:
                    if version_column_name in version["altertable"][table_name]["addcolumn"]:
                        for attr in version["altertable"][table_name]["addcolumn"][version_column_name]:
                            column[attr] = version["altertable"][table_name]["addcolumn"][version_column_name][attr]

    return column


#### Function to check if the column has been renamed
def get_column_renames(version_list: list, column_name: str, table_name: str, order: Optional[Literal["DESC", "ASC"]] = "DESC"):
    """
    Returns list of objects containing rename data:

    old_name: Column name before renaming
    new_name: Column name after renaming
    rename_version: Version in which the column was renamed (parsed to int)
    """
    rename_data = []

    for version in reversed(version_list) if order == "DESC" else version_list:
        if "altertable" in version:
            if table_name in version["altertable"]:
                v = convert_version_number(version["_id"])

                #### Skip versions that are already processed
                if order == "DESC":
                    if any(r["rename_version"] <= v for r in rename_data):
                        continue
                else:
                    if any(r["rename_version"] >= v for r in rename_data):
                        continue

                if "renamecolumn" in version["altertable"][table_name]:
                    renamecolumn_values = list(version["altertable"][table_name]["renamecolumn"].values())

                    #### If the current column is not the one being renamed, continue
                    if order == "DESC" and not column_name in renamecolumn_values:
                        continue

                    renamecolumn_keys = list(version["altertable"][table_name]["renamecolumn"].keys())

                    #### If the current column is not the one being renamed, continue
                    if order == "ASC" and not column_name in renamecolumn_keys:
                        continue

                    #### Get old or new name based on order
                    if order == "DESC":
                        name = renamecolumn_keys[renamecolumn_values.index(column_name)]
                    else:
                        name = renamecolumn_values[renamecolumn_keys.index(column_name)]

                    rename_data.append({"old_name": name if order == "DESC" else column_name, "new_name": name if order == "ASC" else column_name, "rename_version": v})

                    #### Now recursively call it again with the new column column_name
                    rename_data += get_column_renames(version_list, name, table_name, order)
                    break  ## Break the loop as the current column name does not exist

    return rename_data


def get_column_type(version_list: list, table_name: str, column_name: str):
    """
    Find column type in complete version list. Provided column name should be it's most recent
    """

    column_type = None

    rename_data = get_column_renames(version_list=version_list, column_name=column_name, table_name=table_name)
    version_column_name = min(rename_data, key=lambda x: x["rename_version"])["old_name"] if rename_data else column_name
    for version in version_list:
        #### Get the column name for the current version
        v = convert_version_number(version["_id"])
        for rename in rename_data:
            if v >= rename["rename_version"]:
                version_column_name = rename["new_name"]
                break

        if "createtable" in version:
            if table_name in version["createtable"]:
                if version_column_name in version["createtable"][table_name]:
                    column_type = version["createtable"][table_name][version_column_name]["type"]

        if "altertable" in version:
            if table_name in version["altertable"]:
                if "modifycolumn" in version["altertable"][table_name]:
                    if version_column_name in version["altertable"][table_name]["modifycolumn"]:
                        if "type" in version["altertable"][table_name]["modifycolumn"][version_column_name]:
                            column_type = version["altertable"][table_name]["modifycolumn"][version_column_name]["type"]

    return column_type



# Copyright (C) 2023 Wibo Kuipers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional

from alphadb.utils.common import convert_version_number


def get_primary_key(version_list: list, table_name: str, before_version: Optional[str] = None) -> str | None:
    "Get tables primary key from full version source."

    primary_key = None

    for version in version_list:
        #### Skip if version is after or equal to the "before version"
        if not before_version == None and convert_version_number(before_version) <= convert_version_number(version["_id"]):
            continue

        if "createtable" in version:
            if table_name in version["createtable"]:
                if "primary_key" in version["createtable"][table_name]:
                    primary_key = version["createtable"][table_name]["primary_key"]

        if "altertable" in version:
            if table_name in version["altertable"]:
                if "primary_key" in version["altertable"][table_name]:
                    primary_key = version["altertable"][table_name]["primary_key"]

                #### If the column is dropped, primary key should reset to None
                if "dropcolumn" in version["altertable"][table_name]:
                    if not primary_key == None:
                        for dropcol in version["altertable"][table_name]["dropcolumn"]:
                            if dropcol == primary_key:
                                primary_key = None

    return primary_key
